/*
 * Copyright (c) 2015, Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of Freescale Semiconductor, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-----------------------------------------------------------------------*/
/* Based on Low level disk I/O module skeleton for FatFs  (C)ChaN, 2016  */
/*-----------------------------------------------------------------------*/
#include <string.h>
#include <stdbool.h>

#include "ffconf.h"     /* FatFs configuration options */
#include "diskio.h"     /* FatFs lower layer API */
#include "FreeRTOS.h"   /* FreeRTOS interfaces */

#ifdef RAM_DISK_ENABLE
#include "fsl_ram_disk.h"
#endif

#ifdef USB_DISK_ENABLE
#include "fsl_usb_disk.h"
#endif

#ifdef SD_DISK_ENABLE
#include "fsl_sd_disk.h"
#endif

#ifdef MMC_DISK_ENABLE
#include "fsl_mmc_disk.h"
#endif

#ifdef SDSPI_DISK_ENABLE
#include "fsl_sdspi_disk.h"
#endif

/*-----------------------------------------------------------------------*/
/* Copy an un-aligned buffer                                             */
/*-----------------------------------------------------------------------*/
static BYTE *memcpy_unaligned (
    BYTE *dst,                  /* Copy destination */
    const BYTE *src,            /* Copy source */
    UINT len                    /* Copy length */
)
{
    BYTE *pDst = dst;
    const BYTE *pSrc = src;

    /* As pointers can be unaligned, the copy is done byte-wise */
    while (len--)
    {
        *pDst++ = *pSrc++;
    }

    return dst;
}

/*-----------------------------------------------------------------------*/
/* Allocate an aligned buffer                                            */
/*-----------------------------------------------------------------------*/
static DRESULT alloc_aligned_buffer (
    BYTE pdrv,                  /* Physical drive number to identify the drive */
    UINT sector_count,          /* Number of sectors to allocate */
    const BYTE *unaligned_buff, /* Possibly un-aligned buffer */
    BYTE **aligned_buff,        /* Aligned buffer to be allocated */
    bool copyContent            /* If the buffer content needs to be copied */
)
{
    DRESULT res = RES_OK;

    /* Return in most cases the input buffer */
    *aligned_buff = (BYTE *)unaligned_buff;

    /* Check the input buffer alignment */
    if (((UINT)unaligned_buff & (sizeof(UINT) - 1U)) != 0U)
    {
        UINT sector_size = 0U;

        /* Get the disk sector size */
        res = disk_ioctl(pdrv, GET_SECTOR_SIZE, (void *)&sector_size);

        /* Check returned error */
        if (res == RES_OK)
        {
            /* Check valid sector size */
            if (sector_size != 0U)
            {
                /* Allocate a new buffer via RTOS */
                *aligned_buff = (BYTE *)pvPortMalloc(sector_count * sector_size);

                /* Copy the buffer content if needed */
                if (copyContent)
                {
                    /* Copy the buffer content from the un-aligned buffer */
                    (void)memcpy_unaligned(*aligned_buff, unaligned_buff, sector_count * sector_size);
                }
            }
            else
            {
                /* Invalid sector size */
                res = RES_PARERR;
            }
        }
    }

    return res;
}

/*-----------------------------------------------------------------------*/
/* Free an aligned buffer                                                */
/*-----------------------------------------------------------------------*/
static DRESULT dealloc_aligned_buffer (
    BYTE pdrv,                  /* Physical drive number to identify the drive */
    UINT sector_count,          /* Number of sectors to deallocate */
    const BYTE *unaligned_buff, /* Possibly un-aligned buffer */
    BYTE **aligned_buff,        /* Aligned buffer to be freed */
    bool copyContent            /* If the buffer content needs to be copied */
)
{
    DRESULT res = RES_OK;

    /* Check the input buffer alignment */
    if (((UINT)unaligned_buff & (sizeof(UINT) - 1U)) != 0U)
    {
        /* Copy the buffer content if needed */
        if (copyContent)
        {
            UINT sector_size = 0U;

            /* Get the disk sector size */
            res = disk_ioctl(pdrv, GET_SECTOR_SIZE, (void *)&sector_size);

            /* Check returned error */
            if (res == RES_OK)
            {
                /* Check valid sector size */
                if (sector_size != 0U)
                {
                    /* Copy the buffer content to the un-aligned buffer */
                    (void)memcpy_unaligned((BYTE *)unaligned_buff, *aligned_buff, sector_count * sector_size);
                }
                else
                {
                    /* Invalid sector size */
                    res = RES_PARERR;
                }
            }
        }

        /* The aligned buffer was allocated dynamically via RTOS, and must be freed */
        vPortFree((void *)(*aligned_buff));
    }

    return res;
}

/*-----------------------------------------------------------------------*/
/* Get Drive Status                                                      */
/*-----------------------------------------------------------------------*/
DSTATUS disk_status (
    BYTE pdrv                   /* Physical drive number to identify the drive */
)
{
    DSTATUS stat;

    switch (pdrv)
    {
#ifdef RAM_DISK_ENABLE
        case RAMDISK:
            stat = ram_disk_status(pdrv);
            return stat;
#endif
#ifdef USB_DISK_ENABLE
        case USBDISK:
            stat = USB_HostMsdGetDiskStatus(pdrv);
            return stat;
#endif
#ifdef SD_DISK_ENABLE
        case SDDISK:
            stat = sd_disk_status(pdrv);
            return stat;
#endif
#ifdef MMC_DISK_ENABLE
        case MMCDISK:
            stat = mmc_disk_status(pdrv);
            return stat;
#endif
#ifdef SDSPI_DISK_ENABLE
        case SDSPIDISK:
            stat = sdspi_disk_status(pdrv);
            return stat;
#endif
        default:
            break;
    }

    return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Initialize a Drive                                                    */
/*-----------------------------------------------------------------------*/
DSTATUS disk_initialize (
    BYTE pdrv                   /* Physical drive number to identify the drive */
)
{
    DSTATUS stat;

    switch (pdrv)
    {
#ifdef RAM_DISK_ENABLE
        case RAMDISK:
            stat = ram_disk_initialize(pdrv);
            return stat;
#endif
#ifdef USB_DISK_ENABLE
        case USBDISK:
            stat = USB_HostMsdInitializeDisk(pdrv);
            return stat;
#endif
#ifdef SD_DISK_ENABLE
        case SDDISK:
            stat = sd_disk_initialize(pdrv);
            return stat;
#endif
#ifdef MMC_DISK_ENABLE
        case MMCDISK:
            stat = mmc_disk_initialize(pdrv);
            return stat;
#endif
#ifdef SDSPI_DISK_ENABLE
        case SDSPIDISK:
            stat = sdspi_disk_initialize(pdrv);
            return stat;
#endif
        default:
            break;
    }

    return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/
DRESULT disk_read (
    BYTE pdrv,                  /* Physical drive number to identify the drive */
    BYTE *buff,                 /* Data buffer to store read data */
    DWORD sector,               /* Start sector in LBA */
    UINT count                  /* Number of sectors to read */
)
{
    BYTE *aligned_buff;
    DRESULT res;

    /* Allocate an aligned buffer if needed */
    res = alloc_aligned_buffer(pdrv, count, buff, &aligned_buff, false);

    /* Check if we have an error related to buffer management */
    if (res == RES_OK)
    {
        /* Call the underlying layer */
        switch (pdrv)
        {
#ifdef RAM_DISK_ENABLE
            case RAMDISK:
                res = ram_disk_read(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef USB_DISK_ENABLE
            case USBDISK:
                res = USB_HostMsdReadDisk(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef SD_DISK_ENABLE
            case SDDISK:
                res = sd_disk_read(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef MMC_DISK_ENABLE
            case MMCDISK:
                res = mmc_disk_read(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef SDSPI_DISK_ENABLE
            case SDSPIDISK:
                res = sdspi_disk_read(pdrv, aligned_buff, sector, count);
                break;
#endif
            default:
                break;
        }

        /* Deallocate the aligned buffer if needed */
        if (res == RES_OK)
        {
            /* No error so far, so take as result the return code of the dealloc function */
            res = dealloc_aligned_buffer(pdrv, count, buff, &aligned_buff, true);
        }
        else
        {
            /* Already an error, do not overwrite the existing error, but execute dealloc to avoid memory fragmentation */
            (void)dealloc_aligned_buffer(pdrv, count, buff, &aligned_buff, true);
        }
    }

    return res;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/
DRESULT disk_write (
    BYTE pdrv,                  /* Physical drive number to identify the drive */
    const BYTE *buff,           /* Data to be written */
    DWORD sector,               /* Start sector in LBA */
    UINT count                  /* Number of sectors to write */
)
{
    BYTE *aligned_buff;
    DRESULT res;

    /* Allocate an aligned buffer if needed */
    res = alloc_aligned_buffer(pdrv, count, buff, &aligned_buff, true);

    /* Check if we have an error related to buffer management */
    if (res == RES_OK)
    {
        /* Call the underlying layer */
        switch (pdrv)
        {
#ifdef RAM_DISK_ENABLE
            case RAMDISK:
                res = ram_disk_write(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef USB_DISK_ENABLE
            case USBDISK:
                res = USB_HostMsdWriteDisk(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef SD_DISK_ENABLE
            case SDDISK:
                res = sd_disk_write(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef MMC_DISK_ENABLE
            case MMCDISK:
                res = mmc_disk_write(pdrv, aligned_buff, sector, count);
                break;
#endif
#ifdef SDSPI_DISK_ENABLE
            case SDSPIDISK:
                res = sdspi_disk_write(pdrv, aligned_buff, sector, count);
                break;
#endif
            default:
                break;
        }

        /* Deallocate the aligned buffer if needed */
        if (res == RES_OK)
        {
            /* No error so far, so take as result the return code of the dealloc function */
            res = dealloc_aligned_buffer(pdrv, count, buff, &aligned_buff, false);
        }
        else
        {
            /* Already an error, do not overwrite the existing error, but execute dealloc to avoid memory fragmentation */
            (void)dealloc_aligned_buffer(pdrv, count, buff, &aligned_buff, false);
        }
    }

    return res;
}

/*-----------------------------------------------------------------------*/
/* Disk IOCTL                                                            */
/*-----------------------------------------------------------------------*/
DRESULT disk_ioctl (
    BYTE pdrv,                  /* Physical drive number to identify the drive */
    BYTE cmd,                   /* Control code */
    void *buff                  /* Buffer to send/receive control data */
)
{
    DRESULT res;

    switch (pdrv)
    {
#ifdef RAM_DISK_ENABLE
        case RAMDISK:
            res = ram_disk_ioctl(pdrv, cmd, buff);
            return res;
#endif
#ifdef USB_DISK_ENABLE
        case USBDISK:
            res = USB_HostMsdIoctlDisk(pdrv, cmd, buff);
            return res;
#endif
#ifdef SD_DISK_ENABLE
        case SDDISK:
            res = sd_disk_ioctl(pdrv, cmd, buff);
            return res;
#endif
#ifdef MMC_DISK_ENABLE
        case MMCDISK:
            res = mmc_disk_ioctl(pdrv, cmd, buff);
            return res;
#endif
#ifdef SDSPI_DISK_ENABLE
        case SDSPIDISK:
            res = sdspi_disk_ioctl(pdrv, cmd, buff);
            return res;
#endif
        default:
            break;
    }

    return RES_PARERR;
}

/*-----------------------------------------------------------------------*/
/* Get the time from RTC                                                 */
/*-----------------------------------------------------------------------*/
__weak DWORD get_fattime(void)
{
    return 0;
}
