/*
 * C
 *
 * Copyright 2015-2016 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <LLNET_SSL_SOCKET_impl.h>
#include <LLNET_SSL_SOCKET_wolfssl.h>
#include <LLNET_SSL_CONSTANTS.h>
#include "LLNET_CONSTANTS.h"
#include "LLNET_CHANNEL_impl.h"
#include "LLDISPATCH_CONSTANTS.h"
#include "LLDISPATCH_EVENT.h"
#include <wolfssl/ssl.h>
#include <wolfssl/internal.h>
#include <sys/socket.h>
#include "stdio.h"
#include <LLNET_SSL_ERRORS.h>
#include <LLNET_SSL_verifyCallback.h>

#ifdef __cplusplus
	extern "C" {
#endif

/* external functions */
extern void LLNET_SSL_MEM_initialize(void);
extern int32_t LLNET_SSL_TranslateReturnCode(int32_t wolfSSL_error);
extern int32_t LLNET_SSL_VERIFY_verifyCallback(int32_t previousVerifySuccess, WOLFSSL_X509_STORE_CTX* store);
		
int32_t LLNET_SSL_SOCKET_IMPL_initialize(){
	LLNET_SSL_DEBUG_TRACE("%s()\n", __func__);
	LLNET_SSL_MEM_initialize();
	int32_t ret = wolfSSL_Init();
	#ifdef LLNET_SSL_DEBUG
		wolfSSL_Debugging_ON();
	#endif
	return LLNET_SSL_TranslateReturnCode(ret);
}

int32_t LLNET_SSL_SOCKET_IMPL_create(int32_t context, int32_t fd, uint8_t autoclose, int8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, fd=%d)\n", __func__, context, fd);
	WOLFSSL* ssl;
	WOLFSSL_CTX* ctx = (WOLFSSL_CTX*)context;

    /* create new SSL session */
	ssl = wolfSSL_new(ctx);
	if(ssl != NULL){
		if(wolfSSL_set_fd(ssl, fd) != SSL_SUCCESS){
			wolfSSL_free(ssl);
			return J_CREATE_SSL_ERROR; //error
		}else{
			//enable peer verification
			wolfSSL_set_verify((WOLFSSL*)ssl, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, LLNET_SSL_VERIFY_verifyCallback);
			//create and init verify callback context
			struct VERIFY_CALLBACK_CTX* cbCtx = (struct VERIFY_CALLBACK_CTX*)XMALLOC(sizeof(struct VERIFY_CALLBACK_CTX), ctx->heap, DYNAMIC_TYPE_TMP_BUFFER);
			cbCtx->ssl = ssl;
			cbCtx->error = J_SSL_NO_ERROR;
			//set verify callback context
			wolfSSL_SetCertCbCtx(ssl, cbCtx);
		}
	}else{
		return J_CREATE_SSL_ERROR; //error
	}

	LLNET_SSL_DEBUG_TRACE("%s(context=%d, fd=%d) ssl=%p \n", __func__, context, fd, ssl);
	return (int32_t)ssl;
}


int32_t LLNET_SSL_SOCKET_IMPL_close(int32_t ssl, int32_t fd, uint8_t autoclose, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, autoclose=%d, retry=%d)\n", __func__, ssl, fd, autoclose, retry);
	WOLFSSL* wssl = (WOLFSSL*)ssl;

	if(!retry){
		//set non-blocking mode
		int32_t origflags = LLNET_SSL_SOCKET_set_non_blocking(ssl, fd);
		if(origflags >= 0){
			//send close notify
			SendAlert(wssl, alert_warning, close_notify);
			//reset non-blocking mode
			wolfSSL_set_using_nonblock((WOLFSSL*)ssl, 0);
			fcntl(fd, F_SETFL, origflags);
		}
	}

	if(!autoclose){
		//the close of the underlying socket is not requested
		//read close_notify alert to clear input stream
		int8_t buffer[1];
		int32_t res = LLNET_SSL_SOCKET_IMPL_read(ssl, fd, buffer, 0, 1, retry);

		if(res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT || (res >= 0 && !wssl->options.closeNotify)){
			//read operation does not failed and the close notify alert is not yet received
			return J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
		}
	}
    
    return J_SSL_NO_ERROR;
}

int32_t LLNET_SSL_SOCKET_IMPL_freeSSL(int32_t sslID, uint8_t retry)
{
  LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, retry=%d) free SSL\n", __func__, sslID, retry);
  WOLFSSL* pst_ssl_socket = (WOLFSSL*)sslID;	
  
  //free the verify callback context
  XFREE(pst_ssl_socket->verifyCbCtx, pst_ssl_socket->heap, DYNAMIC_TYPE_TMP_BUFFER);
  //free SSL
  wolfSSL_free(pst_ssl_socket);
  
  return J_SSL_NO_ERROR;
}

int32_t LLNET_SSL_SOCKET_set_non_blocking(int32_t ssl, int32_t fd){
	wolfSSL_set_using_nonblock((WOLFSSL*)ssl, 1);
	int32_t origflags = fcntl(fd, F_GETFL, 0);
	if (origflags == -1) {
		return J_SOCKET_ERROR;
	}
	//set non-blocking mode
	int32_t flags = origflags | O_NONBLOCK;
	if (fcntl(fd, F_SETFL, flags) == -1) {
		return J_SOCKET_ERROR;
	}
	return origflags;
}

int32_t LLNET_SSL_SOCKET_asyncOperation(int32_t fd, int32_t operation, uint8_t retry)
{
		int32_t javaThreadID = SNI_getCurrentJavaThreadID();
		int32_t timeout = LLNET_CHANNEL_IMPL_getOption(fd, CPNET_SO_TIMEOUT, 0);
		if(timeout < 0){
			return J_SOCKET_ERROR;
		}

		if((timeout != 0) && retry){
			//The socket has been configured with a timeout
			//AND there is no available data
			//AND this is the second read (retry==true)
			//THEN timeout has been reached during the first read
			return J_SOCKET_TIMEOUT;
		}
		uint8_t added = LLDISPATCH_EVENT_addBlockingRequest(fd, javaThreadID, operation, timeout);
		if(added){
			// request added in the queue
			LLNET_SSL_DEBUG_TRACE("Native Code Has Blocked Without Result\n");
			return J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
		}
		// requests queue limit reached
		LLNET_SSL_DEBUG_TRACE("Blocking Request Queue Limit Reached\n");
		return J_BLOCKING_QUEUE_LIMIT_REACHED;
}

int32_t LLNET_SSL_SOCKET_IMPL_initialClientHandShake(int32_t ssl, int32_t fd, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, retry=%d)\n", __func__, ssl, fd, retry);

	//set non-blocking mode
	int32_t origflags = LLNET_SSL_SOCKET_set_non_blocking(ssl, fd);
	if(origflags < 0){
		return J_SOCKET_ERROR;
	}

	//initiates handshake in non-blocking mode
	int32_t ret = wolfSSL_connect((WOLFSSL*)ssl);

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, retry=%d) ret=%d\n", __func__, ssl, fd, retry, ret);

	//reset non-blocking mode
	wolfSSL_set_using_nonblock((WOLFSSL*)ssl, 0);
	//restore original flags
	if (fcntl(fd, F_SETFL, origflags) == -1) {
		return J_SOCKET_ERROR;
	}

	if(ret != SSL_SUCCESS){
		
		int32_t err = wolfSSL_get_error((WOLFSSL*)ssl, ret);

		if(err == SSL_ERROR_WANT_READ){
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_READ, retry);
		}
		else if(err == SSL_ERROR_WANT_WRITE){
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_WRITE, retry);
		}
		//check verify callback error
		struct VERIFY_CALLBACK_CTX * cbCtx = (struct VERIFY_CALLBACK_CTX *)((WOLFSSL*)ssl)->verifyCbCtx;
		if(cbCtx != NULL && cbCtx->error != J_SSL_NO_ERROR){
			//verify callback failed, return the callback error
			return cbCtx->error; 
		}
		return LLNET_SSL_TranslateReturnCode(err);
	}
	return J_SSL_NO_ERROR;
}

int32_t LLNET_SSL_SOCKET_IMPL_read(int32_t ssl, int32_t fd, int8_t* buffer, int32_t offset, int32_t length, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d)\n", __func__, ssl, fd, offset, length, retry);
	//set non-blocking mode
	int32_t origflags = LLNET_SSL_SOCKET_set_non_blocking(ssl, fd);
	if(origflags < 0){
		return J_SOCKET_ERROR;
	}

	//non-blocking read
	int ret = wolfSSL_read((WOLFSSL*)ssl, buffer+offset, length);

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) bytes read=%d\n", __func__, ssl, fd, offset, length, retry, ret);

	//reset non-blocking mode
	wolfSSL_set_using_nonblock((WOLFSSL*)ssl, 0);
	//restore original flags
	if (fcntl(fd, F_SETFL, origflags) == -1) {
		return J_SOCKET_ERROR;
	}

	if(ret < 0){
		int32_t err = wolfSSL_get_error((WOLFSSL*)ssl, ret);
		if(err == SSL_ERROR_WANT_READ){
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_READ, retry);
		}
		else if(err == SSL_ERROR_WANT_WRITE){
			//read operation can also cause write operation when the peer requests a re-negotiation
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_WRITE, retry);
		}
		return LLNET_SSL_TranslateReturnCode(err);
	}
	
	if(ret == 0){
		//end-of-file
		return J_EOF;
	}
	
	return ret;
	
}


int32_t LLNET_SSL_SOCKET_IMPL_write(int32_t ssl, int32_t fd, int8_t* buffer, int32_t offset, int32_t length, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d)\n", __func__, ssl, fd, offset, length, retry);
	//set non-blocking mode
	int32_t origflags = LLNET_SSL_SOCKET_set_non_blocking(ssl, fd);
	if(origflags < 0){
		return J_SOCKET_ERROR;
	}

	//non-blocking read
	int ret = wolfSSL_write((WOLFSSL*)ssl, buffer+offset, length);

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) written bytes=%d\n", __func__, ssl, fd, offset, length, retry, ret);

	//reset non-blocking mode
	wolfSSL_set_using_nonblock((WOLFSSL*)ssl, 0);
	//restore original flags
	if (fcntl(fd, F_SETFL, origflags) == -1) {
		return J_SOCKET_ERROR;
	}


	if(ret < 0){
		int32_t err = wolfSSL_get_error((WOLFSSL*)ssl, ret);
		if(err == SSL_ERROR_WANT_READ){
			//write operation can also cause read operation when the peer requests a re-negotiation
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_READ, retry);
		}
		else if(err == SSL_ERROR_WANT_WRITE){
			 return LLNET_SSL_SOCKET_asyncOperation(fd, LLDISPATCH_CONSTANTS_REQUEST_WRITE, retry);
		}
		return LLNET_SSL_TranslateReturnCode(err);
	}
	
	if(ret == 0){
		return J_CONNECTION_RESET;
	}
	
	return ret;
}

int32_t LLNET_SSL_SOCKET_IMPL_available(int32_t ssl, uint8_t retry){
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d)\n", __func__, ssl);
	int32_t ret =  wolfSSL_pending((WOLFSSL*)ssl);
	if(ret < 0){
		return J_UNKNOWN_ERROR;
	}
	return ret;
}

#ifdef __cplusplus
	}
#endif
