/*
 * C
 *
 * Copyright 2015 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/* Includes ------------------------------------------------------------------*/
#include "bsp_util.h"
#include "fsl_lcdc.h"
#include "FreeRTOS.h"
#include "semphr.h"
#include "fsl_debug_console.h"
#include "fsl_dma.h"
#include "LLDISPLAY_impl.h"
#include "LLDISPLAY_intern.h"
#include "framerate.h"

/* Definitions ---------------------------------------------------------------*/
/* LCD related macros */
#define APP_LCD LCD
#define LCD_PANEL_CLK 9000000
#define LCD_PPL 480
#define LCD_HSW 2
#define LCD_HFP 8
#define LCD_HBP 43
#define LCD_LPP 272
#define LCD_VSW 10
#define LCD_VFP 4
#define LCD_VBP 12
#define LCD_POL_FLAGS (kLCDC_InvertVsyncPolarity | kLCDC_InvertHsyncPolarity)
#define IMG_HEIGHT LCD_LPP
#define IMG_WIDTH LCD_PPL
#define LCD_INPUT_CLK_FREQ CLOCK_GetFreq(kCLOCK_LCD)
#define APP_LCD_IRQHandler LCD_IRQHandler
#define APP_LCD_IRQn LCD_IRQn
#define IMG_SIZE (IMG_WIDTH * IMG_HEIGHT * APP_BIT_PER_PIXEL / 8)

/* Buffer related macros */
#define BUFFER_SIZE         (IMG_WIDTH * IMG_HEIGHT * APP_BIT_PER_PIXEL / 8)
#define BACK_BUFFER         (DISPLAY_MEM_START)
#define FRAME_BUFFER        (BACK_BUFFER + BUFFER_SIZE)

/* Global -------------------------------------------------------------------*/
/* Declare MicroEJ Display buffer */
BSP_DECLARE_BUFFER(DISPLAY_MEM)

/* DMA handle */
static dma_handle_t dma_handle;

/* DMA transfer configuration */
static dma_transfer_config_t dma_transfer_config;

/* List of chained DMA descriptors, one for each display line (except for the first line, which handled by dma_transfer_config) */
#if defined(__ICCARM__)
#pragma data_alignment = 16U
static dma_descriptor_t dma_descriptors[IMG_HEIGHT - 1] = {0};
#elif defined(__CC_ARM)
static dma_descriptor_t __attribute__((aligned(16U))) dma_descriptors[IMG_HEIGHT - 1] = {0};
#elif defined(__GNUC__)
static dma_descriptor_t __attribute__((aligned(16U))) dma_descriptors[IMG_HEIGHT - 1] = {0};
#endif

/* Synchronization semaphore */
static SemaphoreHandle_t refresh_sem;

/* Variable used for LCD interrupt synchronization with a requested buffer update */
static volatile bool trigger_buffer_update;

/* Private APIs -------------------------------------------------------------*/
static void LLDISPLAY_switch_done(void)
{
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

    /* Give the semaphore to wakeup the caller task */
    xSemaphoreGiveFromISR(refresh_sem, &xHigherPriorityTaskWoken);

    /* Switch tasks if necessary. */
    if (xHigherPriorityTaskWoken != pdFALSE)
    {
        portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
    }
}

static void DMA_Callback(struct _dma_handle *handle, void *userData, bool transferDone, uint32_t intmode)
{
    /* Check DMA transfer status */
    if (transferDone)
    {
        /* UI switch done */
        LLDISPLAY_switch_done();
    }
}

static void LLDISPLAY_StartDMATransfer(void)
{
    /* Start DMA transfer */
    DMA_StartTransfer(&dma_handle);
}

static uint32_t LLDISPLAY_ComputeDMATransferWidth(uint32_t src, uint32_t dest, uint32_t transfer_length)
{
    uint32_t byte_width;

    /* Compute the byte width of the DMA transfer */
    if (((transfer_length & 0x03) == 0x00) && ((src & 0x03) == 0x00) && ((dest & 0x03) == 0x00))
    {
        byte_width = sizeof(uint32_t);
    }
    else if (((transfer_length & 0x01) == 0x00) && ((src & 0x01) == 0x00) && ((dest & 0x01) == 0x00))
    {
        byte_width = sizeof(uint16_t);
    }
    else
    {
        byte_width = sizeof(uint8_t);
    }

    return byte_width;
}

/* This function takes into consideration that full lines will be updated (see implementation of LLDISPLAY_IMPL_flush) */
static void LLDISPLAY_SetupDMATransfer(uint32_t src, uint32_t dest, uint32_t num_transfers)
{
    uint32_t transfer_length = IMG_WIDTH * (APP_BIT_PER_PIXEL / 8); /* One transfer per display line by default */
    void * next_desc = NULL;
    uint32_t byte_width;

    /* Compute the byte width of the DMA transfer */
    byte_width = LLDISPLAY_ComputeDMATransferWidth(src, dest, transfer_length);

    /* Check how many transfers we have to perform */
    if (num_transfers > 1)
    {
        next_desc = &dma_descriptors[0];
    }

    /* Create the first DMA descriptor */
    DMA_PrepareTransfer(&dma_transfer_config,
                        (void *)src,
                        (void *)dest,
                        byte_width,
                        transfer_length,
                        kDMA_MemoryToMemory,
                        next_desc);

    /* Check if a DMA chained operation needs to be set-up */
    if (num_transfers > 1)
    {
        /* Configure the transfer configuration to not enable an interrupt after each transfer */
        dma_transfer_config.xfercfg.intA = false;
        dma_transfer_config.xfercfg.swtrig = true;

        /*
         * Create chained DMA descriptors:
         * - The first one will be dma_transfer_config
         * - The second one will be dma_descriptors[0]
         * - ...
         * - The last one will be dma_descriptors[num_transfers - 2]
         */
        for (uint32_t idx = 0; idx < (num_transfers - 1); idx++)
        {
            DMA_CreateDescriptor(&dma_descriptors[idx],
                                 &dma_transfer_config.xfercfg,
                                 (void *)(src + (idx + 1) * transfer_length),
                                 (void *)(dest + (idx + 1) * transfer_length),
                                 &dma_descriptors[idx + 1]);
        }

        /* Configure the last DMA descriptor to enable an interrupt at the end of the transfer and to disable the descriptor reload */
        dma_descriptors[num_transfers - 2].xfercfg |= DMA_CHANNEL_XFERCFG_SETINTA_MASK;
        dma_descriptors[num_transfers - 2].xfercfg &= ~DMA_CHANNEL_XFERCFG_RELOAD_MASK;
        dma_descriptors[num_transfers - 2].linkToNextDesc = NULL;
    }

    /* Submit DMA transfer */
    (void)DMA_SubmitTransfer(&dma_handle, &dma_transfer_config);
}

void APP_LCD_IRQHandler(void)
{
    uint32_t intStatus = LCDC_GetEnabledInterruptsPendingStatus(APP_LCD);

    /* Clear interrupt status */
    LCDC_ClearInterruptsStatus(APP_LCD, intStatus);

    /* If an update is requested */
    if ((intStatus & kLCDC_BaseAddrUpdateInterrupt) && trigger_buffer_update)
    {
        /* Update trigger variable */
        trigger_buffer_update = false;

        /* Start DMA transfer */
        LLDISPLAY_StartDMATransfer();
    }
    __DSB();
}

/* Public APIs --------------------------------------------------------------*/
void LLDISPLAY_IMPL_initialize(void)
{
    lcdc_config_t lcdConfig;

    /* Create synchronization sempahore */
    vSemaphoreCreateBinary(refresh_sem);
    xSemaphoreTake(refresh_sem, 0);

    /* Update trigger variable */
    trigger_buffer_update = false;

    /* Get default LCD configuration */
    LCDC_GetDefaultConfig(&lcdConfig);

    /* Configure LCD panel */
    lcdConfig.panelClock_Hz = LCD_PANEL_CLK;
    lcdConfig.ppl = LCD_PPL;
    lcdConfig.hsw = LCD_HSW;
    lcdConfig.hfp = LCD_HFP;
    lcdConfig.hbp = LCD_HBP;
    lcdConfig.lpp = LCD_LPP;
    lcdConfig.vsw = LCD_VSW;
    lcdConfig.vfp = LCD_VFP;
    lcdConfig.vbp = LCD_VBP;
    lcdConfig.polarityFlags = LCD_POL_FLAGS;
    lcdConfig.upperPanelAddr = (uint32_t)BACK_BUFFER;
    lcdConfig.bpp = kLCDC_16BPP565;
    lcdConfig.display = kLCDC_DisplayTFT;
    lcdConfig.swapRedBlue = true;

    /* Initialize LCD panel */
    if (LCDC_Init(APP_LCD, &lcdConfig, LCD_INPUT_CLK_FREQ) != kStatus_Success)
    {
        /* Error during initialization */
        return;
    }

    /* Configure LCD DMA watermark level */
    LCDC_SetDmaFifoWatermarkLevel(APP_LCD, kLCDC_DmaFifoEightEmptyLocations);

    /* Configure and enable LCD interrupts */
    LCDC_EnableInterrupts(APP_LCD, kLCDC_BaseAddrUpdateInterrupt);
    NVIC_SetPriority(APP_LCD_IRQn, 4);
    NVIC_EnableIRQ(APP_LCD_IRQn);

    /* Start the LCD panel */
    LCDC_Start(APP_LCD);
    LCDC_PowerUp(APP_LCD);

    /* Initialize DMA channel used for display buffer copy */
    DMA_Init(DMA0);
    DMA_EnableChannel(DMA0, 0);
    DMA_CreateHandle(&dma_handle, DMA0, 0);
    DMA_SetCallback(&dma_handle, DMA_Callback, NULL);

    /* Configure and enable DMA channel interrupt */
    NVIC_SetPriority(DMA0_IRQn, 3);
    NVIC_EnableIRQ(DMA0_IRQn);
}

int32_t LLDISPLAY_IMPL_getWidth(void)
{
    return IMG_WIDTH;
}

int32_t LLDISPLAY_IMPL_getHeight(void)
{
    return IMG_HEIGHT;
}

void LLDISPLAY_IMPL_synchronize(void)
{
    /* Wait for the update to be finished */
    xSemaphoreTake(refresh_sem, portMAX_DELAY);
}

int32_t LLDISPLAY_IMPL_getGraphicsBufferAddress(void)
{
	return (int32_t)BACK_BUFFER;
}

int32_t LLDISPLAY_IMPL_flush(int32_t addr, int32_t xmin, int32_t ymin, int32_t xmax, int32_t ymax)
{
    int32_t new_back_buffer = (addr == BACK_BUFFER) ? FRAME_BUFFER : BACK_BUFFER;
    uint32_t offset = ((uint32_t)ymin * IMG_WIDTH * APP_BIT_PER_PIXEL / 8);

#ifdef FRAMERATE_ENABLED
    /* Increment framerate */
    framerate_increment();
#endif

    /* Setup DMA transfer */
    LLDISPLAY_SetupDMATransfer(addr + offset, new_back_buffer + offset, (uint32_t)ymax - (uint32_t)ymin + 1);

    /* Set new LCD panel address */
    LCDC_SetPanelAddr(APP_LCD, kLCDC_UpperPanel, (uint32_t)addr);

    /* Update trigger variable */
    trigger_buffer_update = true;

    return new_back_buffer;
}
